/**********************************************************************************/
/*                                                                                */
/* File: draw.cpp                                                                 */
/* Author: bkenwright@xbdev.net                                                   */
/* URL: www.xbdev.net                                                             */
/* Date: 10-11-2005                                                               */
/*                                                                                */
/**********************************************************************************/
/*
    Nothing exciting here - this is where we get our triangle data, and put it
	into a set of structures, D3DXVECTOR3 for the vertice information and
	tridata for an array of triangle index values.  tridata class is defined
	at the top of progmesh and is just 3 integer values, for the 3 index
	values into the vertex data;

*/
/**********************************************************************************/

#pragma comment(lib, "D3d8.lib")
#pragma comment(lib, "D3dx8.lib")
#include <d3dx8.h>


#include "progmesh.h"
#include "rabdata.h"


extern IDirect3DDevice8	* g_pd3dDevice; // defined at the top of main.cpp

int   render_num ;   // Number of vertices to draw with
                     // Set it initially to the same value as the number of vertices
                     // we load, e.g. RABBIT_VERTEX_NUM, then we can dynamically
					 // decrement it so we render less and less verts to
					 // see the effects of different mesh sizes.



D3DXVECTOR3 vert[RABBIT_VERTEX_NUM];       // global list of vertices
tridata tri[RABBIT_TRIANGLE_NUM];       // global list of triangles

int collapse_map[RABBIT_TRIANGLE_NUM * 3];  // to which neighbor each vertex collapses


// Note that the use of the Map() function and the collapse_map
// list isn't part of the polygon reduction algorithm.
// We just set up this system here in this module
// so that we could retrieve the model at any desired vertex count.
// Therefore if this part of the program confuses you, then
// dont worry about it.  It might help to look over the progmesh.cpp
// module first.

//       Map()
//
// When the model is rendered using a maximum of mx vertices
// then it is vertices 0 through mx-1 that are used.  
// We are able to do this because the vertex list 
// gets sorted according to the collapse order.
// The Map() routine takes a vertex number 'a' and the
// maximum number of vertices 'mx' and returns the 
// appropriate vertex in the range 0 to mx-1.
// When 'a' is greater than 'mx' the Map() routine
// follows the chain of edge collapses until a vertex
// within the limit is reached.
//   An example to make this clear: assume there is
//   a triangle with vertices 1, 3 and 12.  But when
//   rendering the model we limit ourselves to 10 vertices.
//   In that case we find out how vertex 12 was removed
//   by the polygon reduction algorithm.  i.e. which
//   edge was collapsed.  Lets say that vertex 12 was collapsed
//   to vertex number 7.  This number would have been stored
//   in the collapse_map array (i.e. collapse_map[12]==7).
//   Since vertex 7 is in range (less than max of 10) we
//   will want to render the triangle 1,3,7.  
//   Pretend now that we want to limit ourselves to 5 vertices.
//   and vertex 7 was collapsed to vertex 3 
//   (i.e. collapse_map[7]==3).  Then triangle 1,3,12 would now be
//   triangle 1,3,3.  i.e. this polygon was removed by the
//   progressive mesh polygon reduction algorithm by the time
//   it had gotten down to 5 vertices.
//   No need to draw a one dimensional polygon. :-)
int Map(int a,int mx) 
{
	if(mx<=0) return 0; // Simple error checking, can't have negative array values
	while(a>=mx) {  
		a=collapse_map[a];
	}
	return a;
}//End Map(..)



/**********************************************************************************/

////////////////////////////////////////////////////////////////////////////////////
//                                                                                //
//  RenderSingleTriangle(...)                                                     //
//  Takes 3 vertice values, I've used 3 pointers.  Not very efficent, as we       //
//  are doing a triangle by triangle rendering, creating and releasing resources  //
//  over and over again, but its simple and doens't complicate things!            //
//  As this is a tutorial :)                                                      //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
void RenderSingleTriangle(D3DXVECTOR3 *pvA, D3DXVECTOR3 *pvB, D3DXVECTOR3 *pvC,
						  DWORD dwColour)
{
	struct stVertex
	{
		D3DXVECTOR3 pos;  // D3DFVF_XYZ
		DWORD colour;     // D3DFVF_DIFFUSE
	};

 
	UINT FVF_Vertex = (D3DFVF_XYZ | D3DFVF_DIFFUSE);

	IDirect3DVertexBuffer8 * DX_vb;
	
	g_pd3dDevice->CreateVertexBuffer( 3*sizeof(stVertex), 0, FVF_Vertex, D3DPOOL_MANAGED, &DX_vb );

	stVertex *pVB;
    DX_vb->Lock(0,0, (BYTE**)&pVB, 0);
	pVB[0].pos    = *pvA;
	pVB[0].colour = dwColour;

	pVB[1].pos    = *pvB;
	pVB[1].colour = dwColour;

	pVB[2].pos    = *pvC;
	pVB[2].colour = dwColour;

    DX_vb->Unlock();


	g_pd3dDevice->SetStreamSource(0, DX_vb, sizeof(stVertex));
    g_pd3dDevice->SetVertexShader(FVF_Vertex);

    g_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 1);

	DX_vb->Release();
	
}//End RenderSingleTriangle(..)

////////////////////////////////////////////////////////////////////////////////////
//                                                                                //
//  DrawModelTriangles(...)                                                       //
//  Basically uses our collaps_map(..) data we generated from the progressive     //
//  mesh algorith and produces our mesh, it draws the number of triangles         //
//  depending on the size of render_num, so if render_num is the number of        //
//  of vertices if set to the how many there really is it will draw all the       //
//  model...if if wet render_num to 50% of the vertice values it draws half the   //
//  mesh with half the values.                                                    //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////

void DrawModelTriangles() 
{
	int num_tris = RABBIT_TRIANGLE_NUM;

	// Keep count of how many triangles we pass along to our screen for rendering
	// Its only used for informative purposes or debugging etc.
	int renderpolycount=0;
	int i=0;
	for(i=0;i<num_tris;i++) {
		int p0= Map(tri[i].v[0],render_num);
		int p1= Map(tri[i].v[1],render_num);
		int p2= Map(tri[i].v[2],render_num);
		// note:  serious optimization opportunity here,
		//  by sorting the triangles the following "continue" 
		//  could have been made into a "break" statement.
		if(p0==p1 || p1==p2 || p2==p0) continue;
		renderpolycount++;

		D3DXVECTOR3 v0,v1,v2; 
		v0 = vert[p0];
		v1 = vert[p1];
		v2 = vert[p2];


		// Take 3 vertices, which of course 3 points make up a triangle and
		// pass them along to our directx function and render it.
		RenderSingleTriangle((D3DXVECTOR3*)&v0, // --|
							(D3DXVECTOR3*)&v1,  // --+-- 3 triangle points
							(D3DXVECTOR3*)&v2,  // --|
							0xff00ffff);        // Colour Triangle
	}
}



////////////////////////////////////////////////////////////////////////////////////
//                                                                                //
//  GetRabbitData(...)                                                            //
//  Not to complicated, simply takes our data that we've got in rabdata.h/cpp     //
//  and copies it across into our data structure so we can pass it to our         //
//  progressive mesh function                                                     //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
void GetRabbitData()
{
	// Copy the geometry from the arrays of data in rabdata.cpp into
	// the vert and tri lists which we send to the reduction routine
	int i;
	for(i=0;i<RABBIT_VERTEX_NUM;i++) 
	{
		float *vp=rabbit_vertices[i];
		vert[i] = D3DXVECTOR3(vp[0],vp[1],vp[2]);
	}
	for(i=0;i<RABBIT_TRIANGLE_NUM;i++) 
	{
		tridata td;
		td.v[0]=rabbit_triangles[i][0];
		td.v[1]=rabbit_triangles[i][1];
		td.v[2]=rabbit_triangles[i][2];
		
		tri[i] = td;
	}
	render_num=RABBIT_VERTEX_NUM; // by default lets use all the model to render
}


////////////////////////////////////////////////////////////////////////////////////
//                                                                                //
//  InitModel()                                                                   //
//  Our Progressive Mesh function is called here!  We get our mesh data, then     //
//  call our BuildProgressiveMesh(..) function so it builds our collapse_map      //
//  array for us.  Then we can use collapse_map to render our various levels      //
//  of detail mesh, ranging from 100% there to 1% etc very blocky version.        //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
void InitModel() 
{
	int num_verts = RABBIT_VERTEX_NUM;
	int num_tris  = RABBIT_TRIANGLE_NUM;


	// Get our vertice/triangle indices values and put them into an array
	GetRabbitData();  
	

	// Run our progressive mesh algorithm in progmesh.cpp/h which returns the
	// filled in array collapse_map which we can use to render the various
	// levels of detail of mesh
	BuildProgressiveMesh(num_verts,
						 num_tris,
						 vert,tri,
						 collapse_map);
	
}//End InitModel(..)


void DeInitModel() 
{
}


////////////////////////////////////////////////////////////////////////////////////
//                                                                                //
//  RenderModel(...)                                                              //
//  Our recursive render loop!  Called over and over again from our main.cpp      //
//  file.  So from here we can set any directx values and render to the screen    //
//  using our vertice values - of course we call DrawModelTriangles(..)           //
//  which renders our mesh.                                                       //
//  Each time round, render_num is decremented and hence each time round our      //
//  mesh is rendered with less and less triangles.                                //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
void RenderModel() 
{
	
	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	//g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );

	static int max_num = render_num;

	render_num--;
	if( render_num < 10 )
		render_num = max_num;


	 DrawModelTriangles();

}// End RenderModel(..)

