/**********************************************************************************/
/*                                                                                */
/*  File: progmesh.cpp                                                            */
/*  Author: bkenwright@xbdev.net                                                  */
/*  URL: www.xbdev.net                                                            */
/*  Date: 10-11-2005                                                              */
/*                                                                                */
/**********************************************************************************/
/*
	Progressive Mesh Algorithm (or Level Of Details LOD)

	NOTE 'Progressive', as the algorithm produces a list of values into our vertice
	array so that we can create any size of mesh we need quickly just using the
	result from our algorith

	SLOOWWW!!  At the moment, due to the 'Collapse(..)' function using lots of
	recursive loops and of course checking all the vertices each time!  This slows
	the creation code right down!  This can be optimised with a bit of thinking
	and work!  Once its created its very fast!

	The algorithm has been kept to its simplest form!  A real treat is to take it
	to the next level by optimising it and adding in some further algorithms
	to interpolate textures and other vertice values.

	The data passed to the algorithm is a list of vertices and an list of triangles
	that index our vertice list.

	To keep the triangle structure, a triangle structure is used for the 3 values
	that make up each face.

	In our algorithms we do numerous things to create our progressive mesh!  We
	create neighboring vertice lists, neighboring face lists etc, so that as our
	mesh gets smaller we can decide which vertice is worth the least amount.

	See progmesh.h for further details.
*/
/**********************************************************************************/


#include <stdio.h>  
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <windows.h>


#include <d3dx8.h> // D3DXVECTOR3, D3DXVec3Length(..)


#include "progmesh.h"



/**********************************************************************************/

// Forward declaration of our two classes                                        //
//
class xTriangle;
class xVertex;


/**********************************************************************************/

////////////////////////////////////////////////////////////////////////////////////
//
// Our various vertices and triangle index values need to be put into a more 
// complex structure so they can hold futher information.  Such as each vertice 
// needs to know which are its neighboring vertices and also each triangle face 
// needs to know its face normal value etc.
//
////////////////////////////////////////////////////////////////////////////////////

class xTriangle 
{
  public:
	xVertex *        vertex[3]; // the 3 points that make this tri
	D3DXVECTOR3      normal;    // unit D3DXVECTOR3 othogonal to this face
	                 xTriangle(xVertex *v0,xVertex *v1,xVertex *v2); // constructor
	                
	void             ComputeNormal();

	void			 ReplaceVertex(xVertex*vold, xVertex*vnew);

	int              HasVertex(xVertex *v);
};


class xVertex 
{
  public:
	D3DXVECTOR3      position; // location of point in euclidean space
	int              id;       // place of aVertex in original list

	#define MAX_ADJ_VERTS 700
	int num_neighbors;
	xVertex*        neighbor[MAX_ADJ_VERTS];   // adjacent vertices

	#define MAX_ADJ_FACES 700
	int num_faces;
	xTriangle*      face[MAX_ADJ_FACES];     // adjacent cTriangles

	float            objdist;  // cached cost of collapsing edge
	xVertex *        collapse; // candidate aVertex for collapse
	               
					 xVertex(D3DXVECTOR3 v,int _id); // Constructor
};



/**********************************************************************************/




xTriangle::xTriangle(xVertex *v0,xVertex *v1,xVertex *v2)
{
	assert(v0!=v1 && v1!=v2 && v2!=v0);
	vertex[0]=v0;
	vertex[1]=v1;
	vertex[2]=v2;
	ComputeNormal();

	for(int i=0;i<3;i++) 
	{
		vertex[i]->face[ vertex[i]->num_faces++ ] = this;
		for(int j=0;j<3;j++) if(i!=j) 
		{
			vertex[i]->neighbor[ vertex[i]->num_neighbors++ ] = vertex[j];
		}
	}
}//End xTriangle(..) Constructor


int xTriangle::HasVertex(xVertex *v) 
{
	return (v==vertex[0] ||v==vertex[1] || v==vertex[2]);
}//End HasVertex(..)


void xTriangle::ComputeNormal()
{
	D3DXVECTOR3 v0=vertex[0]->position;
	D3DXVECTOR3 v1=vertex[1]->position;
	D3DXVECTOR3 v2=vertex[2]->position;
	
	D3DXVECTOR3 vA = (v1-v0);
	D3DXVECTOR3 vB = (v2-v1);

	D3DXVec3Cross(&normal, &vA, &vB);

	float magnitude = D3DXVec3Length(&normal);
	if(magnitude==0)return;
	D3DXVec3Normalize(&normal,&normal);
}//End ComputeNormal(..)


void xTriangle::ReplaceVertex(xVertex *vold, xVertex *vnew) 
{
	for(int i=0; i<3; i++)
		if( vertex[i]==vold) vertex[i]=vnew;

	ComputeNormal();
}//End ReplaceVertex(..)


/**********************************************************************************/


xVertex::xVertex(D3DXVECTOR3 v,int _id) 
{
	position =v;
	id=_id;

	num_neighbors = 0;
	num_faces = 0;
}




/**********************************************************************************/


float ComputeEdgeCollapseCost(xVertex *u, xVertex *v) 
{
	if( u==NULL) return 0;
	if( v==NULL) return 0;
	// if we collapse edge uv by moving u to v then how 
	// much different will the model change, i.e. how much "error".
	// Texture, aVertex normal, and border aVertex code was removed
	// to keep this demo as simple as possible.
	// The method of determining cost was designed in order 
	// to exploit small and coplanar regions for
	// effective polygon reduction.
	// Is is possible to add some checks here to see if "folds"
	// would be generated.  i.e. normal of a remaining face gets
	// flipped.  I never seemed to run into this problem and
	// therefore never added code to detect this case.
	int i;
	D3DXVECTOR3 vlen = (v->position) - (u->position);
	float edgelength = D3DXVec3Length( &(vlen) );
	float curvature=0;


	// Find the "sides" cTriangles that are on the edge uv
	int num_sides = 0;
	//xTriangle * sides[RABBIT_TRIANGLE_NUM];
	// We have just defined a large temp buffer for the max number of tris
	// shouldn't be any where near 1000!, but can be made more dynamic at
	// a later date - such as count how many sides first then allocate memory
	// for example
	#define MAX_TRIS 1000
	xTriangle ** sides = new xTriangle* [MAX_TRIS];

	for(i=0; i<u->num_faces; i++)
	{
		if( u->face[i]==NULL) continue;

		if( u->face[i]->HasVertex(v) )
		{
			sides[num_sides++] = u->face[i];
		}

		// Buffer overflow!
		assert(num_sides < MAX_TRIS);
	}

	// Use the triangle facing most away from the sides 
	// to determine our curvature term
	for(i=0; i<u->num_faces; i++)
	{
		float mincurv=1.0f;
		if(u->face[i]==NULL) continue;

		for(int j=0; j<num_sides; j++)
		{
			if(sides[j]==NULL) continue;
			// use dot product of face normals. defined in D3DXVECTOR3
			float dotprod = D3DXVec3Dot( &u->face[i]->normal , &sides[j]->normal );
			mincurv = min(mincurv,(1-dotprod)/2.0f);
		}
		curvature = max(curvature,mincurv);
	}

	delete[] sides;

	// the more coplanar the lower the curvature term   
	return edgelength * curvature;
}

/**********************************************************************************/


void ComputeEdgeCostAtaVertex(xVertex *v) 
{
	if( v==NULL ) return;
	// compute the edge collapse cost for all edges that start
	// from aVertex v.  Since we are only interested in reducing
	// the object by selecting the min cost edge at each step, we
	// only cache the cost of the least cost edge at this aVertex
	// (in member variable collapse) as well as the value of the 
	// cost (in member variable objdist).
	
	//count how many neighbours are 0!
	int v_num_neighbors = 0;
	for(int j=0; j<v->num_neighbors; j++)
	{
		if( v->neighbor[j] != NULL ) v_num_neighbors++;
	}


	if( v_num_neighbors == 0 )
	{
		// v doesn't have neighbors so it costs nothing to collapse
		v->collapse=NULL;
		v->objdist=-0.01f;
		return;
	}
	v->objdist = 1000000;
	v->collapse=NULL;

	// search all neighboring edges for "least cost" edge
	for(int i=0; i<v->num_neighbors; i++)
	{
		if( v->neighbor[i] == NULL) continue;

		float dist;
		dist = ComputeEdgeCollapseCost(v, v->neighbor[i]);
		if( dist < v->objdist )
		{
			v->collapse = v->neighbor[i];
			v->objdist  = dist;
		}
	}
}

/**********************************************************************************/


void ComputeAllEdgeCollapseCosts(int num_verts, xVertex ** verts) 
{
	// For all the edges, compute the difference it would make
	// to the model if it was collapsed.  The least of these
	// per aVertex is cached in each aVertex object.
	for(int i=0;i<num_verts;i++) 
	{
		ComputeEdgeCostAtaVertex(verts[i]);
	}
}

/**********************************************************************************/

void calc_faces(int num_verts, int num_tris,
				xVertex ** verts, xTriangle ** tris)
{

	for(int j=0; j<num_verts; j++)
		if( verts[j]!=NULL )
			verts[j]->num_faces = 0;

	for(int i=0; i<num_verts; i++)
	{
		if(verts[i]== NULL) continue;

		for(int j=0; j<num_tris; j++)
		{
			if( tris[j]==NULL ) continue;

			if( tris[j]->HasVertex(verts[i]) )
				verts[i]->face[ verts[i]->num_faces++ ]=tris[j];
		}

	}
	
}

// Call calc_faces(..) first, so that when we call this function
// all the vertice face values have been set, else it won't be able
// to use them to determine the neighboring vertice values.
void calc_neighbors(int num_verts, int num_tris,
					xVertex ** verts, xTriangle ** tris)
{

	for(int m=0; m<num_verts; m++)
		if( verts[m]!=NULL )
			verts[m]->num_neighbors=0;

	for(int i=0; i<num_verts; i++)
	{
		if( verts[i]== NULL) continue;

		for( int j=0; j<verts[i]->num_faces; j++)
		{
			if( verts[i]->face[j]==NULL) continue;

			for(int k=0; k<3; k++)
			{
				if( verts[i]->face[j]->vertex[k] != verts[i] )
				{
					// We should check there is only one of them!
						verts[i]->neighbor[ verts[i]->num_neighbors++ ] = 
							                      verts[i]->face[j]->vertex[k];
				}
			}
		}
	}
	
}

void CalcAllNormals(int num_tris, xTriangle ** tris )
{
	for(int i=0; i<num_tris; i++)
	{
		if( tris[i]==NULL ) continue;

		tris[i]->ComputeNormal();
	}
}

void calc_all_faces_normals( int num_verts, int num_tris,
					         xVertex ** verts, xTriangle ** tris)
{
	calc_faces(num_verts, num_tris, verts, tris);
	calc_neighbors(num_verts, num_tris, verts, tris);
	CalcAllNormals(num_tris, tris);
}


/**********************************************************************************/

////////////////////////////////////////////////////////////////////////////////////
//                                                                                //
//   Collapse(..)                                                                 //
//   Its a lot complicate than it really is.                                      //
//   Takes some time to break it down...all the loops and things make it messy    //
//   Simply take v and u input values.  Then we go through all our vertex values  //
//   in our array and put v onto u!  So after our function finished there     //
//   should be no existences of u in our list of triangles or vertices.           //
//                                                                                //
//   So we have to check all the neighbor arrays, face arrays, and our large      //
//   array of triangles and vertices.                                             //
//                                                                                //
//   Then once we have replaced them, we need to recalculate all the weight       //
//   values for each vertice, so we know which vertice we can remove next now     //
//   that this vertice has gone.                                                  //
//                                                                                //
//   Its worth actually looking at a sketch of what happens when we move v onto   //
//   u.  As a number of things are looking at, for example, which neighbor        //
//   vertices are effects, also which faces are destroyed and need removing etc.  //
//                                                                                //
////////////////////////////////////////////////////////////////////////////////////
void Collapse(int num_verts, int num_tris,
			  xVertex **verts, xTriangle** tris,
			  
			  xVertex *u,xVertex *v)
{
	// Collapse the edge uv by moving aVertex u onto v
	// Actually remove tris on uv, then update tris that
	// have u to have v, and then remove u.
	
	// u is removed
	// v replaces u

	int i;

	if(v==NULL) 
	{
		// u is a vertex all by itself so just delete it
		for(i=num_verts-1; i>=0; i--)
		{
			if( verts[i]== NULL) continue;
			
			if( verts[i]== u )
			{
				delete verts[i];
				verts[i] = NULL;
				return;
			}
		}
		//return;
	}


	int num_tmp=0;
	xVertex ** tmp = new xVertex* [num_verts]; 
	// make tmp a list of all the neighbors of u
	for(i=0;i<u->num_neighbors;i++) 
	{
		if( u->neighbor[i]==NULL) continue;
		tmp[num_tmp++] = u->neighbor[i];
	}


	//Go through all our array of vertices
	for(i=num_verts-1; i>=0; i--)
	{
		if( verts[i]== NULL) continue;
		
		if( verts[i]== u )
		{
			delete verts[i];
			verts[i] = NULL;
		}
	}


	
	//Go through all our array of triangles and remove
	//any dead or no-longer needed triangles that have u and v
	for(i=num_tris-1; i>=0; i--)
	{
		if( tris[i] == NULL ) continue;
			
		if( tris[i]->HasVertex(u) &&
			tris[i]->HasVertex(v))
		{
			delete tris[i];
			tris[i] = NULL;
		}
	}

	//Go through all our array of triangles and any triangles that
	//use vertice u, well we just replace that vertice with v.
	for(i=num_tris-1; i>=0; i--)
	{
		if( tris[i] == NULL ) continue;

		tris[i]->ReplaceVertex(u,v);
	}
	

	// Serious optimisation fault!  We don't really need to do all the
	// triangles and vertices!  Would be better to work out with neighboring
	// faces and vertices are effected and only recalcuate there values
	// greatly improving the speed.

	calc_all_faces_normals(num_verts, num_tris,  //****/// OPTIMISATION ISSUE!! //*****//
		                   verts, tris); 

	// Recompute the edge collapse costs for neighboring vertices

	//	ComputeAllEdgeCollapseCosts(num_verts);	
	
	// Optimisation!
	// Its usually easier to see this visually, but we are only removing
	// 1 point in our vertice mesh!  Hence the only values to be effected
	// by this change, are the values which are around it!  Which are
	// the neighboring vertices
	for(i=0;i<num_tmp;i++) {
		ComputeEdgeCostAtaVertex(tmp[i]);
	}

	delete[] tmp;

}//End Collapse(..)



/**********************************************************************************/

void AddVertex(int num_verts, int num_tris, D3DXVECTOR3 * vert,
			   xVertex ** vert_out)
{
	for(int i=0;i<num_verts;i++) 
	{
		vert_out[i] = new xVertex(vert[i],i);
	}
}//End AddVertex(..)


void AddFaces(int num_tris, tridata * tri, 
			  xVertex ** verts, 
			  xTriangle ** tri_out)
{
	for(int i=0;i<num_tris;i++) {
			tri_out[i] =new xTriangle(
					      verts[tri[i].v[0]],
					      verts[tri[i].v[1]],
					      verts[tri[i].v[2]] );
	}
}//End AddFaces(..)


/**********************************************************************************/


xVertex *MinimumCostEdge(int num_verts, xVertex ** verts)
{
	// Find the edge that when collapsed will affect model the least.
	// This funtion actually returns a aVertex, the second aVertex
	// of the edge (collapse candidate) is stored in the aVertex data.
	// Serious optimization opportunity here: this function currently
	// does a sequential search through an unsorted list :-(
	// Our algorithm could be O(n*lg(n)) instead of O(n*n)
	xVertex *mn = verts[0];

	// Get the first vertex from our array of vertex's that hasn't
	// been deleted
	for(int j=0; j<num_verts; j++)
	{
		if( verts[j]!= NULL )
		{
			mn = verts[j];
			break;
		}
	}
	

	// Compare our vertex value that we have with every other vertex objdist
	// to make sure we have the smallest value
	for(int i=0;i<num_verts;i++) 
	{
		if( verts[i] == NULL ) continue;

		if(verts[i]->objdist < mn->objdist )
		{
			mn = verts[i];
		}
	}

	return mn;
}//End MinimumCostEdge(..)



/**********************************************************************************/


void ProgressiveMesh(int num_verts,
					 int num_tris,
					 D3DXVECTOR3 * vert, tridata * tri, 
                     int * map,         // Array of int's the same size as number of vertices
					 int * permutation) // Array of int's the same size as number of vertices
{
	xVertex   ** verticesA	 = new xVertex*	 [num_verts];
	xTriangle ** cTrianglesA = new xTriangle* [num_tris];



	AddVertex(num_verts, num_tris, vert,
		      verticesA);  // put input data into our data structures
	AddFaces(num_tris, tri,
			 verticesA, cTrianglesA);

	//calc_faces();
	//calc_neighbors();
	//CalcAllNormals();

	ComputeAllEdgeCollapseCosts(num_verts, verticesA); // cache all edge collapse costs

	int ii = num_verts;
	while(ii > 0) 
	{
		// get the next aVertex to collapse
		xVertex *mn = MinimumCostEdge(num_verts, verticesA);
		// keep track of this aVertex, i.e. the collapse ordering
		permutation[mn->id]=ii-1;
		// keep track of aVertex to which we collapse to
		map[ii-1] = (mn->collapse)?mn->collapse->id:-1;
		// Collapse this edge
		Collapse(num_verts, num_tris,
			     verticesA, cTrianglesA,
				 mn,mn->collapse);
		ii--;
		
	}
	
	// reorder the map list based on the collapse ordering
	for(int i=0;i<num_verts;i++) 
	{
		map[i] = (map[i]==-1)?0:permutation[map[i]];
	}
	
	// The caller of this function should reorder their vertices
	// according to the returned "permutation".

	// Release Resources
	for(int i=0; i<num_verts; i++) delete verticesA[i];
	for(int j=0; j<num_tris; j++) delete cTrianglesA[j];

	delete[] verticesA;
	delete[] cTrianglesA;

}//End ProgressiveMesh(..)


/**********************************************************************************/


void PermuteVertices(int num_verts, int num_tris,
					 D3DXVECTOR3 * vert, tridata * tri, 
					 int * permutation) 
{
	D3DXVECTOR3 * temp_list = new D3DXVECTOR3[num_verts];

	int i;

	for(i=0;i<num_verts;i++) 
	{
		temp_list[i] = vert[i];
	}

	for(i=0;i<num_verts;i++) {
		vert[permutation[i]]=temp_list[i];
	}

	// update the changes in the entries in the triangle list
	for(i=0;i<num_tris;i++) {
		for(int j=0;j<3;j++) {
			tri[i].v[j] = permutation[tri[i].v[j]];
		}
	}

	delete[] temp_list;

}//End PermuteVertices(..)


/**********************************************************************************/

void BuildProgressiveMesh(	int num_verts,
							int num_tris,
							D3DXVECTOR3 * vert, tridata * tri, 
							int * collapse_map  )
{
	int * permutation = new int[num_verts];

	ProgressiveMesh(num_verts,
					num_tris,
		            vert,tri,
					collapse_map,
					permutation);


	PermuteVertices(num_verts, num_tris, 
					vert,tri,
					permutation);

	delete[] permutation;

}//End BuildProgressiveMesh





